(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2021 *)

(** A module that defines pretty printers for the various values for
   the types we have defined so far *)

open Ast

let pp_var fmt x = Format.pp_print_string fmt x
let pp_loc fmt loc = Format.fprintf fmt "@@%i" loc
let pp_fname fmt x = Format.pp_print_string fmt x

let pp_v fmt =
  let open Format in
  function
  | Unit -> pp_print_string fmt "()"
  | Bool b -> pp_print_bool fmt b
  | Int n -> pp_print_int fmt n
  | Loc l -> pp_loc fmt l

let pp_imm fmt = function Var x -> pp_var fmt x | Val v -> pp_v fmt v

let pp_binop fmt =
  let open Format in
  function
  | AND -> pp_print_string fmt "&&"
  | OR -> pp_print_string fmt "||"
  | LE -> pp_print_string fmt "<="
  | LT -> pp_print_string fmt "<"
  | GE -> pp_print_string fmt ">="
  | GT -> pp_print_string fmt ">"
  | NEQ -> pp_print_string fmt "<>"
  | EQ -> pp_print_string fmt "=="
  | PLUS -> pp_print_string fmt "+"
  | MINUS -> pp_print_string fmt "-"
  | MULT -> pp_print_string fmt "*"

let pp_args fmt args =
  let open Format in
  pp_print_list ~pp_sep:(fun fmt () -> fprintf fmt ",@ ") pp_imm fmt args

let pp_formal_args fmt args =
  let open Format in
  pp_print_list ~pp_sep:(fun fmt () -> fprintf fmt ",@ ") pp_var fmt args

let pp_a fmt =
  let open Format in
  function
  | Imm i -> pp_imm fmt i
  | Binop (op, i1, i2) -> fprintf fmt "%a %a %a" pp_imm i1 pp_binop op pp_imm i2
  | NewRef i -> fprintf fmt "ref %a" pp_imm i
  | GetRef x -> fprintf fmt "!%a" pp_var x
  | SetRef (x, i) -> fprintf fmt "%a := %a" pp_var x pp_imm i
  | Call (f, args) -> fprintf fmt "@[%a(%a)@]" pp_fname f pp_args args

let rec pp_cmd fmt =
  let open Format in
  function
  | Return a -> pp_a fmt a
  | AssignVar (x, a) -> fprintf fmt "@[<v 0>%a = @[%a@]@]" pp_var x pp_a a
  | If (cond, cmds1, cmds2) ->
      fprintf fmt
        "@[<v 0>if %a@ @[<v 2>then {@ @[%a@]@]@ }@ @[<v 2>else {@ @[%a@]@]@ }@]"
        pp_imm cond pp_cmds cmds1 pp_cmds cmds2
  | While (cond, cmds) ->
      fprintf fmt "@[<v 2>while %a {@ @[%a@]@,@]}" pp_imm cond pp_cmds cmds

and pp_cmds fmt cmds =
  let open Format in
  fprintf fmt "@[%a@]"
    (pp_print_list ~pp_sep:(fun fmt () -> fprintf fmt ";@ ") pp_cmd)
    cmds

let pp_func fmt (args, t) =
  Format.fprintf fmt "fun@ (@[%a@])@ ->@ %a" pp_formal_args args pp_cmds t

let pp_program fmt defs =
  let open Format in
  fprintf fmt "{@[<v 1>";
  fprintf fmt "[@[";
  pp_print_list
    ~pp_sep:(fun fmt () -> fprintf fmt ";@ ")
    (fun fmt (f, func) ->
      fprintf fmt "@[<v 2>%a =@ @[%a@]@]" pp_fname f pp_func func)
    fmt defs;
  fprintf fmt "@,]@]@ ";
  fprintf fmt "@,}@]"
