(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright © Inria 2021 *)

%{
    open Ast
%}

%token EOF LBRACE RBRACE LPAR RPAR
%token FUN WHILE REF COLONEQ EQ LE LT GE GT NEQ EQEQ PLUS MINUS MULT IF THEN ELSE COMMA SEMI BANG LAND LOR
%token<bool> BOOL
%token<string> ID
%token<int> INT

%start<program> program

%%

var:
| x = ID
{ x }

value:
| LPAR RPAR
{ Unit }
| b = BOOL
{ Bool b }
| i = INT
{ Int i }

imm:
| v = value
{ Val v }
| x = ID
{ Var x }

args:
| l = delimited(LPAR, separated_list(COMMA, imm), RPAR)
{ l }

formal_args:
| l = delimited(LPAR, separated_list(COMMA, var), RPAR)
{ l }

binop:
| LAND
{ AND }
| LOR
{ OR }
| LE
{ LE }
| LT
{ LT }
| GE
{ GE }
| GT
{ GT }
| NEQ
{ NEQ }
| EQEQ
{ EQ }
| PLUS
{ Ast.PLUS }
| MINUS
{ Ast.MINUS }
| MULT
{ Ast.MULT }

atom:
| i = imm
{ Imm i }
| i1 = imm op = binop i2 = imm
{ Binop(op, i1, i2) }
| REF i = imm
{ NewRef i }
| BANG x = var
{ GetRef x }
| x = var COLONEQ i = imm
{ SetRef (x, i) }
| p = var xs = args
{ Call(p, xs) }

atomic_cmd:
| a = atom
{ Return a }
| x = var EQ a = atom
{ AssignVar(x, a)  }
| IF cond = imm THEN LBRACE cmds1 = cmds RBRACE ELSE LBRACE cmds2 = cmds RBRACE
| IF cond = imm THEN cmds1 = cmds ELSE LBRACE cmds2 = cmds RBRACE
{ If(cond, cmds1, cmds2) }
| WHILE cond = imm LBRACE cmds = cmds RBRACE
{ While (cond, cmds) }

cmd:
| cmd = atomic_cmd
{ cmd }
| IF cond = imm THEN LBRACE cmds1 = cmds RBRACE ELSE cmd2 = atomic_cmd
| IF cond = imm THEN cmds1 = cmds ELSE cmd2 = atomic_cmd
{ If(cond, cmds1, [cmd2]) }
| WHILE cond = imm cmd = atomic_cmd
{ While (cond, [cmd]) }

cmds:
| l = separated_nonempty_list(SEMI, cmd)
{ l }

fundef:
| FUN p = var xs = formal_args EQ LBRACE cmds = cmds RBRACE
{ (p, (xs, cmds)) }

fundefs:
| defs = fundef+
{ defs }

program:
| defs = fundefs EOF
{ defs }
